# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pathlib

import click

from .. import constants, paranoid, report_github_issue_message
from ..encoding import decode_unicode_chunks
from ..smali import SmaliField, register

logger = logging.getLogger(__name__)


@click.group(name="helpers", help="Helper commands")
def cli():
    pass


@cli.command(help="Extracts the strings from a paranoid obfuscated APK")
@click.argument("target", type=click.Path(exists=True, file_okay=False))
def extract_strings(target: str):
    target_directory = pathlib.Path(target)

    potential_get_string_methods = []
    potential_obfuscated_string_arrays = []
    for smali_file in target_directory.rglob("*.smali"):
        with open(smali_file, "r") as f:
            smali_parser = paranoid.ParanoidSmaliParser(filename=str(smali_file))

            for line_num, line in enumerate(f):
                smali_parser.update(line, line_num)

            # Add potential get string methods
            for method, data in smali_parser.methods.items():
                if (
                    data["consts"] == constants.PARANOID_GET_STRING_CONST_SIGNATURE
                    and method.arguments == constants.PARANOID_GET_STRING_ARGUMENTS
                    and method.return_type == constants.PARANOID_GET_STRING_RETURN_TYPE
                ):
                    potential_get_string_methods.append((method, data["sget_objects"]))

            # Add potential obfuscated string arrays
            for field, data in smali_parser.fields.items():
                if field.type == "[Ljava/lang/String;":
                    potential_obfuscated_string_arrays.append((field, data["value"]))

    # Check if only one method is found
    if len(potential_get_string_methods) != 1:
        logger.error("Found more than one potential get string method")
        logger.error("This is not supported yet")
        return

    get_string_method, get_string_fields = potential_get_string_methods[0]
    get_string_field: SmaliField = get_string_fields[0]

    # Check if only one field is found
    if len(get_string_fields) != 1:
        logger.error("Found more than one potential obfuscated string array")
        logger.error("This is not supported yet")
        return

    # Extract the string chunks
    chunks = []
    for field, value in potential_obfuscated_string_arrays:
        if field.class_name == get_string_field.class_name and field.name == get_string_field.name:
            chunks = value

    # Check if the chunks are found
    if not chunks:
        logger.error("No chunks found")
        return

    logger.debug(f"Method: {get_string_method}")
    logger.debug(f"Field: {get_string_field}")
    # logger.debug("Chunks:")
    # logger.debug(chunks)

    # Find all the deobfuscation values
    deobfuscation_values = []
    for smali_file in target_directory.rglob("*.smali"):
        with open(smali_file, "r") as f:
            smali_parser = paranoid.ParanoidSmaliParser(filename=str(smali_file), target_method=get_string_method)

            for line_num, line in enumerate(f):
                try:
                    smali_parser.update(line, line_num)
                except paranoid.ParanoidSmaliParserError as e:
                    # Ignore Parameters are not supported
                    if e.args[0] == "Parameters are not supported":
                        logger.warning(f"{smali_file}:{line_num+1}: Detected unsupported method call")
                        continue

                    # Log and raise the error
                    logger.error(report_github_issue_message(str(e)))
                    raise e

            deobfuscation_values.extend(smali_parser.state["calls_to_target_method"])

    logger.debug("Deobfuscation values:")
    logger.debug(deobfuscation_values)

    # Decode the chunks
    chunks = decode_unicode_chunks(chunks)

    # Get the raw value
    values = []
    for x in deobfuscation_values:
        if isinstance(x, register.SmaliRegisterConst):
            values.append(x.value)

    # Deobfuscate the strings
    for x in values:
        print(f"[{x:x}]:{paranoid.deobfuscate_string(x, chunks, True)}")



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any


class SmaliField:
    def __init__(
        self,
        name: str,
        type: str,
        modifiers: list[str] = [],
        class_name: str | None = None,
        init_value: str | None = None,
    ):
        self.name = name
        self.type = type
        self.modifiers = modifiers
        self.class_name = class_name
        self.init_value = init_value

    @classmethod
    def from_string(cls, data: str, class_name: str | None = None):
        if not data.startswith(".field"):
            raise ValueError("Invalid field string")

        parts = data.split()

        # Check if the field has value
        init_value = None
        if "=" in parts:
            init_value = parts[-1]
            parts = parts[: parts.index("=")]

        field_modifiers = parts[1:-1]
        field_name, field_type = parts[-1].split(":")

        return cls(field_name, field_type, field_modifiers, class_name, init_value)

    def to_smali(self):
        if self.init_value:
            return f".field {' '.join(self.modifiers)} {self.name}:{self.type} = {self.init_value}"

        return f".field {' '.join(self.modifiers)} {self.name}:{self.type}"

    def __repr__(self):
        return f"SmaliField(name={self.name}, type={self.type}, modifiers={self.modifiers}, class_name={self.class_name}, init_value={self.init_value})"

    def __eq__(self, other: Any):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return (
            self.name == other.name
            and self.type == other.type
            and self.modifiers == other.modifiers
            and self.class_name == other.class_name
            and self.init_value == other.init_value
        )

    def __hash__(self):
        return hash((self.name, self.type, tuple(self.modifiers), self.class_name, self.init_value))


class SmaliMethod:
    def __init__(
        self,
        method: str,
        arguments: list[str] = [],
        return_type: str = "V",
        modifiers: list[str] = [],
        class_name: str | None = None,
    ):
        self.method = method
        self.arguments = arguments
        self.return_type = return_type
        self.modifiers = modifiers
        self.class_name = class_name

    @staticmethod
    def parse_arguments_string(data: str):
        arguments: list[str] = []

        is_in_class = False

        tmp = ""
        for c in data:
            tmp += c

            # Start of a fully qualified class name
            if c == "L":
                is_in_class = True
                continue

            # End of a fully qualified class name
            if is_in_class and c == ";":
                is_in_class = False
                arguments.append(tmp)
                tmp = ""
                continue

            # Primitive type
            if not is_in_class and c in "VZBSCIJFD":
                arguments.append(tmp)
                tmp = ""
                continue

        return arguments

    @staticmethod
    def parse_method_signature(data: str):
        method_name = data.split("(")[0]
        method_return_type = data.split(")")[1]

        # Parse arguments
        method_arguments_raw = data.split("(")[1].split(")")[0]
        method_arguments = SmaliMethod.parse_arguments_string(method_arguments_raw)

        return method_name, method_arguments, method_return_type

    @classmethod
    def from_string(cls, data: str, class_name: str | None = None):
        if not data.startswith(".method"):
            raise ValueError("Invalid method string")

        parts = data.split()
        method_signature = parts[-1]
        method_modifiers = parts[1:-1]

        method_name, method_arguments, method_return_type = SmaliMethod.parse_method_signature(method_signature)

        return cls(method_name, method_arguments, method_return_type, method_modifiers, class_name)

    def to_smali(self):
        return f".method {' '.join(self.modifiers)} {self.method}({''.join(self.arguments)}){self.return_type}"

    def __repr__(self):
        return f"SmaliMethod(method={self.method}, arguments={self.arguments}, return_type={self.return_type}, modifiers={self.modifiers}, class_name={self.class_name})"

    def __eq__(self, other: Any):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return (
            self.method == other.method
            and self.arguments == other.arguments
            and self.return_type == other.return_type
            and self.modifiers == other.modifiers
            and self.class_name == other.class_name
        )

    def __hash__(self):
        return hash((self.method, tuple(self.arguments), self.return_type, tuple(self.modifiers), self.class_name))



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

PARANOID_GET_STRING_CONST_SIGNATURE = [
    4294967295,
    33,
    7109453100751455733,
    28,
    -3808689974395783757,
    32,
    65535,
    16,
    -65536,
    0,
]
PARANOID_GET_STRING_ARGUMENTS = ["J"]
PARANOID_GET_STRING_RETURN_TYPE = "Ljava/lang/String;"



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List


def decode_unicode(data: str):
    """
    Decodes a string containing Unicode escape sequences into its corresponding Unicode characters.

    Args:
        data (str): The input string containing Unicode escape sequences.

    Returns:
        str: The decoded string with Unicode characters.

    Examples:
        >>> decode_unicode("\\\\u30b9\\\\u30fc\\\\u30ab")
        "\u30b9\u30fc\u30ab"
    """
    return data.encode().decode("unicode-escape")


def encode_unicode(data: str):
    """
    Encodes a string into a Unicode escape sequence.

    Args:
        data (str): The input string to be encoded.

    Returns:
        str: The encoded string with Unicode escape sequences.

    Examples:
        >>> encode_unicode("\u30b9\u30fc\u30ab")
        "\\\\u30b9\\\\u30fc\\\\u30ab"
    """
    return data.encode("unicode-escape").decode()


def decode_unicode_chunks(chunks: List[str]):
    """
    Decodes a list of Unicode-encoded strings.

    Args:
        chunks (List[str]): A list of Unicode-encoded strings.

    Returns:
        List[str]: A list of decoded strings.

    Examples:
        >>> decode_unicode_chunks(["\\\\u30b9\\\\u30fc\\\\u30ab"])
        ["\u30b9\u30fc\u30ab"]
    """
    return [decode_unicode(chunk) for chunk in chunks]


def encode_smali_string(data: str | bytes):
    """
    Encodes a given string or bytes into a format suitable for Smali.

    This function takes a string or bytes, converts it to a Unicode string if necessary,
    and then replaces certain characters to ensure the string is properly escaped for Smali.
    Specifically, it replaces double quotes with escaped double quotes, single quotes with
    escaped single quotes, and hexadecimal byte sequences with Unicode escape sequences.

    Args:
        data (str | bytes): The input data to be encoded. It can be either a string or bytes.

    Returns:
        str: The encoded string, with special characters properly escaped for Smali.

    Examples:
        >>> encode_smali_string("\\b")
        "\\\\u0008"
        >>> encode_smali_string(b"test\\"'")
        "test\\\\\\"\\\\'"
    """
    data = data if isinstance(data, str) else bytes(data).decode()
    return encode_unicode(data).replace('"', '\\"').replace("'", "\\'").replace("\\x", "\\u00")



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


def to_int(num: np.integer | int, bit_size: int = 32, signed: bool = True) -> np.integer:
    """
    Convert a number to a numpy integer of a specified bit size and sign.

    Parameters:
        num (int): The number to convert.
        bit_size (int, optional): The bit size of the integer. Must be one of 8, 16, 32, or 64. Default is 32.
        signed (bool, optional): Whether the integer should be signed or unsigned. Default is True.

    Returns:
        np.integer: The converted numpy integer.

    Raises:
        ValueError: If the bit size is not one of 8, 16, 32, or 64.
        ValueError: If the number is out of range for the specified bit size.
    """
    if bit_size not in (8, 16, 32, 64):
        raise ValueError("Invalid bit size. Must be 8, 16, 32, or 64.")

    if not (0 <= num <= 2**bit_size - 1 or -(2 ** (bit_size - 1)) <= num < 2 ** (bit_size - 1)):
        raise ValueError(f"Number out of range for {bit_size}-bit integer.")

    if signed:
        if num < 0:
            if bit_size == 8:
                return np.int8(num)
            elif bit_size == 16:
                return np.int16(num)
            elif bit_size == 32:
                return np.int32(num)
            elif bit_size == 64:
                return np.int64(num)
        else:
            if bit_size == 8:
                return np.uint8(num).view(np.uint8)
            elif bit_size == 16:
                return np.uint16(num).view(np.uint16)
            elif bit_size == 32:
                return np.uint32(num).view(np.uint32)
            elif bit_size == 64:
                return np.uint64(num).view(np.uint64)
    else:
        if num < 0:
            if bit_size == 8:
                return np.int8(num).view(np.uint8)
            elif bit_size == 16:
                return np.int16(num).view(np.uint16)
            elif bit_size == 32:
                return np.int32(num).view(np.uint32)
            elif bit_size == 64:
                return np.int64(num).view(np.uint64)
        else:
            if bit_size == 8:
                return np.uint8(num)
            elif bit_size == 16:
                return np.uint16(num)
            elif bit_size == 32:
                return np.uint32(num)
            elif bit_size == 64:
                return np.uint64(num)



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This class was converted from the following Java code:
#  - https://github.com/MichaelRocks/paranoid/blob/7030aea9aeb8d245c3bea2ef6df20d104d198fce/core/src/main/java/io/michaelrocks/paranoid/RandomHelper.java
#  - https://github.com/LSPosed/LSParanoid/blob/91e1231bc0062d1f90685d739bff9bc50a2b57b0/core/src/main/java/org/lsposed/lsparanoid/RandomHelper.java

import numpy as np

from . import utils


# MurmurHash3 "Mix4" variant
def seed(x: np.int64 | np.integer | int) -> np.int64:
    _x = utils.to_int(x, 64, signed=False)

    with np.errstate(over="ignore"):
        z = (_x ^ (_x >> 33)) * utils.to_int(0x62A9D9ED799705F5, 64)
        y = (z ^ (z >> 28)) * utils.to_int(0xCB24D0A5C88C35B3, 64)

        return y >> 32


def rotl(x: np.int16 | np.integer | int, k: int) -> np.int16:
    _x = utils.to_int(x, 32, signed=False)  # Need to be not signed to get zeros while shifting right

    with np.errstate(over="ignore"):
        return np.int16(_x << k | _x >> (32 - k))


def next(state: np.int64 | np.integer | int) -> np.int64:
    _state = utils.to_int(state, 64, signed=False)

    with np.errstate(over="ignore"):
        s0 = np.int16(_state & 0xFFFF)
        s1 = np.int16((_state >> 16) & 0xFFFF)
        next = s0
        next += s1
        next = rotl(next, 9)
        next += s0

        s1 ^= s0
        s0 = rotl(s0, 13)
        s0 ^= s1
        s0 ^= s1 << 5
        s1 = rotl(s1, 10)

        result = np.int64(next)
        result <<= np.int64(16)
        result |= np.int64(s1)
        result <<= np.int64(16)
        result |= np.int64(s0)
        return result



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Any

# Regex components
# CONST_WIDE_REGEX_PATTERN = r"const-wide(?:/16|/32|/high16)?"
# CONST_REGEX_PATTERN = r"const(?:/4|/16|/high16)?"
# CONST_VALUE_REGEX_PATTERN = r"-?0x[0-9a-f]+"
# REGISTER_REGEX_PATTERN = r"([vp][0-9]+)"
# FULLY_CLASSIFIED_NAME = r"L[a-zA-Z0-9$_\- \u00A0-\u1FFF\u2000-\u200A\u2010-\u2027\u202F\u2030-\uD7FF\uE000-\uFFEF\U00010000-\U0010FFFF/]+;"
# SIMPLE_NAME = r"[a-zA-Z0-9$_\- \u00A0-\u1FFF\u2000-\u200A\u2010-\u2027\u202F\u2030-\uD7FF\uE000-\uFFEF\U00010000-\U0010FFFF]+"
# TYPE_DESCRIPTOR = r"\[*(?:" + FULLY_CLASSIFIED_NAME + r"|[VZBSCIJFD])"
# STATIC_GET_PUT_FIELD_REGEX = r"(" + FULLY_CLASSIFIED_NAME + r")->(" + SIMPLE_NAME + r"):(" + TYPE_DESCRIPTOR + r")"

CONST = re.compile(r"const(?:/4|/16|/high16|-wide(?:/16|/32|/high16)?)?\s+([vp][0-9]+),\s+(-?0x[0-9a-fA-F]+)")  # noqa: E501
CONST_STRING = re.compile(r'const-string\s+([vp][0-9]+),\s+"(.+)"')  # noqa: E501
NEW_ARRAY = re.compile(
    r"new-array\s+([vp][0-9]+),\s+([vp][0-9]+),\s+(\[*(?:L[a-zA-Z0-9$_\- \u00A0-\u1FFF\u2000-\u200A\u2010-\u2027\u202F\u2030-\uD7FF\uE000-\uFFEF\U00010000-\U0010FFFF/]+;|[VZBSCIJFD]))"  # noqa: E501
)
AGET_APUT = re.compile(
    r"a(?:put|get)(?:-(?:wide|object|boolean|byte|char|short))?\s+([vp][0-9]+),\s+([vp][0-9]+),\s+([vp][0-9]+)"  # noqa: E501
)
SGET_SPUT = re.compile(
    r"s(?:put|get)(?:-(?:wide|object|boolean|byte|char|short))?\s+([vp][0-9]+),\s+((L[a-zA-Z0-9$_\- \u00A0-\u1FFF\u2000-\u200A\u2010-\u2027\u202F\u2030-\uD7FF\uE000-\uFFEF\U00010000-\U0010FFFF/]+;)->([a-zA-Z0-9$_\- \u00A0-\u1FFF\u2000-\u200A\u2010-\u2027\u202F\u2030-\uD7FF\uE000-\uFFEF\U00010000-\U0010FFFF]+):(\[*(?:L[a-zA-Z0-9$_\- \u00A0-\u1FFF\u2000-\u200A\u2010-\u2027\u202F\u2030-\uD7FF\uE000-\uFFEF\U00010000-\U0010FFFF/]+;|[VZBSCIJFD])))"  # noqa: E501
)
INVOKE_STATIC = re.compile(
    r"invoke-static\s+{([vp][0-9]+(?:,\s*[vp][0-9]+)*)},\s+(L[a-zA-Z0-9$_\- \u00A0-\u1FFF\u2000-\u200A\u2010-\u2027\u202F\u2030-\uD7FF\uE000-\uFFEF\U00010000-\U0010FFFF/]+;)->([a-zA-Z0-9$_\- \u00A0-\u1FFF\u2000-\u200A\u2010-\u2027\u202F\u2030-\uD7FF\uE000-\uFFEF\U00010000-\U0010FFFF/\[\(\);]+)"  # noqa: E501
)
INVOKE_STATIC_RANGE = re.compile(
    r"invoke-static/range\s*{([vp][0-9]+\s*\.\.\s*[vp][0-9]+)},\s*(L[a-zA-Z0-9$_\- \u00A0-\u1FFF\u2000-\u200A\u2010-\u2027\u202F\u2030-\uD7FF\uE000-\uFFEF\U00010000-\U0010FFFF/]+;)->([a-zA-Z0-9$_\- \u00A0-\u1FFF\u2000-\u200A\u2010-\u2027\u202F\u2030-\uD7FF\uE000-\uFFEF\U00010000-\U0010FFFF/\[\(\);]+)"  # noqa: E501
)
MOVE_RESULT = re.compile(r"move-result(?:-(?:wide|object))?\s+([vp][0-9]+)")  # noqa: E501


class SmaliInstrConst:
    def __init__(self, register: str, value: int, _raw: str | None = None):
        self.register = register
        self.value = value
        self._raw = _raw

    @staticmethod
    def parse(data: str):
        data = data.strip()

        m = CONST.match(data)
        if not m:
            raise ValueError("Invalid const instruction")

        register = m.group(1)
        value = int(m.group(2), 0)

        return SmaliInstrConst(register, value, _raw=data)

    def __eq__(self, other: Any):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return self.register == other.register and self.value == other.value


class SmaliInstrConstString:
    def __init__(self, register: str, value: str, _raw: str | None = None):
        self.register = register
        self.value = value
        self._raw = _raw

    @staticmethod
    def parse(data: str):
        data = data.strip()

        m = CONST_STRING.match(data)
        if not m:
            raise ValueError("Invalid const-string instruction")

        register = m.group(1)
        value = m.group(2)

        return SmaliInstrConstString(register, value, _raw=data)

    def __eq__(self, other: Any):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return self.register == other.register and self.value == other.value


class SmaliInstrInvokeStatic:
    def __init__(self, registers: list[str], class_name: str, method: str, _raw: str | None = None):
        self.registers = registers
        self.class_name = class_name
        self.method = method
        self._raw = _raw

    @staticmethod
    def parse(data: str):
        data = data.strip()

        m = INVOKE_STATIC.match(data)
        if not m:
            raise ValueError("Invalid invoke-static instruction")

        registers = [x.strip() for x in m.group(1).split(",")]
        class_name = m.group(2)
        method = m.group(3)

        return SmaliInstrInvokeStatic(registers, class_name, method, _raw=data)

    def __eq__(self, other: Any):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return self.registers == other.registers and self.class_name == other.class_name and self.method == other.method


class SmaliInstrInvokeStaticRange:
    def __init__(self, registers: list[str], class_name: str, method: str, _raw: str | None = None):
        self.registers = registers
        self.class_name = class_name
        self.method = method
        self._raw = _raw

    @staticmethod
    def parse(data: str):
        data = data.strip()

        m = INVOKE_STATIC_RANGE.match(data)
        if not m:
            raise ValueError("Invalid invoke-static/range instruction")

        # TODO: do we need to include all the possible registers inside the range?
        # For example: {v0 .. v5} -> v0, v1, v2, v3, v4, v5
        registers = [x.strip() for x in m.group(1).split("..")]
        class_name = m.group(2)
        method = m.group(3)

        return SmaliInstrInvokeStaticRange(registers, class_name, method, _raw=data)

    def __eq__(self, other: Any):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return self.registers == other.registers and self.class_name == other.class_name and self.method == other.method


class SmaliInstrNewArray:
    def __init__(self, register: str, size_register: str, type_descriptor: str, _raw: str | None = None):
        self.register = register
        self.size_register = size_register
        self.type_descriptor = type_descriptor
        self._raw = _raw

    @staticmethod
    def parse(data: str):
        data = data.strip()

        m = NEW_ARRAY.match(data)
        if not m:
            raise ValueError("Invalid new-array instruction")

        register = m.group(1)
        size_register = m.group(2)
        type_descriptor = m.group(3)

        return SmaliInstrNewArray(register, size_register, type_descriptor, _raw=data)

    def __eq__(self, other: Any):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return (
            self.register == other.register
            and self.size_register == other.size_register
            and self.type_descriptor == other.type_descriptor
        )


class SmaliInstrAGetAPut:
    def __init__(self, register_dest: str, register_array: str, register_index: str, _raw: str | None = None):
        self.register_dest = register_dest
        self.register_array = register_array
        self.register_index = register_index
        self._raw = _raw

    @staticmethod
    def parse(data: str):
        data = data.strip()

        m = AGET_APUT.match(data)
        if not m:
            raise ValueError("Invalid aget/aput instruction")

        register_dest = m.group(1)
        register_array = m.group(2)
        register_index = m.group(3)

        return SmaliInstrAGetAPut(register_dest, register_array, register_index, _raw=data)

    def __eq__(self, other: Any):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return (
            self.register_dest == other.register_dest
            and self.register_array == other.register_array
            and self.register_index == other.register_index
        )


class SmaliInstrSGetSPut:
    def __init__(self, register_dest: str, class_name: str, field_name: str, field_type: str, _raw: str | None = None):
        self.register_dest = register_dest
        self.class_name = class_name
        self.field_name = field_name
        self.field_type = field_type
        self._raw = _raw

    @staticmethod
    def parse(data: str):
        data = data.strip()

        m = SGET_SPUT.match(data)
        if not m:
            raise ValueError("Invalid sget/sput instruction")

        register_dest = m.group(1)
        # full_field = m.group(2)
        class_name = m.group(3)
        field_name = m.group(4)
        field_type = m.group(5)

        return SmaliInstrSGetSPut(register_dest, class_name, field_name, field_type, _raw=data)

    def __eq__(self, other: Any):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return (
            self.register_dest == other.register_dest
            and self.class_name == other.class_name
            and self.field_name == other.field_name
            and self.field_type == other.field_type
        )


class SmaliInstrMoveResult:
    def __init__(self, register: str, _raw: str | None = None):
        self.register = register
        self._raw = _raw

    @staticmethod
    def parse(data: str):
        data = data.strip()

        m = MOVE_RESULT.match(data)
        if not m:
            raise ValueError("Invalid move-result instruction")

        register = m.group(1)

        return SmaliInstrMoveResult(register, _raw=data)

    def __eq__(self, other: Any):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return self.register == other.register



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This class was converted from the following Java code:
#  - https://github.com/MichaelRocks/paranoid/blob/7030aea9aeb8d245c3bea2ef6df20d104d198fce/core/src/main/java/io/michaelrocks/paranoid/DeobfuscatorHelper.java
#  - https://github.com/LSPosed/LSParanoid/blob/91e1231bc0062d1f90685d739bff9bc50a2b57b0/core/src/main/java/org/lsposed/lsparanoid/DeobfuscatorHelper.java

from typing import List

import numpy as np

from . import RandomHelper, utils

MAX_CHUNK_LENGTH = 0x1FFF


def getString(id: int, chunks: List[str]) -> bytes:
    _id = utils.to_int(id, 64, signed=False)

    with np.errstate(over="ignore"):
        state = RandomHelper.seed(_id & 0xFFFFFFFF)
        state = RandomHelper.next(state)
        low = np.uint64(state) >> 32 & 0xFFFF
        state = RandomHelper.next(state)
        high = np.uint64(state) >> 16 & 0xFFFF0000
        index = (_id >> 32) ^ low ^ high
        state = getCharAt(index, chunks, state)
        length = np.uint64(state) >> 32 & 0xFFFF

        output = []
        for x in range(length):
            state = getCharAt(index + x + 1, chunks, state)
            output.append(np.uint64(state) >> 32 & 0xFFFF)

        return "".join(map(chr, output)).encode("unicode-escape")


def getCharAt(char_index, chunks, state):
    next_state = RandomHelper.next(state)
    chunk = chunks[int(char_index / MAX_CHUNK_LENGTH)]
    char = chunk[int(char_index % MAX_CHUNK_LENGTH)]

    return np.int64(np.uint64(next_state) ^ (np.uint64(ord(char)) << np.uint16(32)))



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
from typing import Any, Dict, List, Literal, TypedDict, overload

from ..smali import SmaliField, SmaliMethod, instructions, register
from . import DeobfuscatorHelper, RandomHelper

# Expose the following classes and functions to the outside world
__all__ = ["DeobfuscatorHelper", "RandomHelper"]

logger = logging.getLogger(__name__)


class SmaliRegisterEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, register.SmaliRegister):
            return o.to_dict()
        return super().default(o)


class ParanoidSmaliParserError(Exception):
    """
    Exception raised for errors encountered by the Paranoid Smali Parser.

    Attributes:
        message (str): The error message.
        extra (Dict[str, Any]): Additional information about the error.

    Methods:
        __str__(): Returns the string representation of the error, including any extra information if available.
    """

    def __init__(self, message: str, extra: Dict[str, Any] = {}):
        super().__init__(message)

        self.extra = extra

    def __str__(self):
        if not self.extra:
            return super().__str__()

        return f"{super().__str__()}\n{json.dumps(self.extra, indent=4, cls=SmaliRegisterEncoder)}"


class ParanoidSmaliParser:
    """
    ParanoidSmaliParser is a class designed to parse and deobfuscate Smali code.

        class_name (str | None): The fully qualified name of the class being parsed.
        fields (Dict[SmaliField, ParanoidSmaliParser.Field]): A dictionary mapping SmaliField instances to their corresponding field data.
        methods (Dict[SmaliMethod, ParanoidSmaliParser.Method]): A dictionary mapping SmaliMethod instances to their corresponding method data.
        target_method (SmaliMethod | None): The target method to look for during parsing.
        _metadata (str): The filename of the Smali file being parsed.
        state (ParanoidSmaliParser.State): The current state of the deobfuscation process.

    Methods:
        __init__(filename: str, target_method: SmaliMethod | None = None):
            Initializes the parser with the given filename and optional target method.

        _reset_state(key_to_reset: str | None = None):
            Resets the parser state to its default values or resets a specific key in the state.

        get_fully_qualified_class_name(line: str) -> str:
            Extracts and returns the fully qualified class name from a Smali class definition line.

        update(line: str):
            Updates the parser state based on the given line of Smali code.
    """

    class Field(TypedDict):
        """
        A TypedDict representing a field with a value and an optional register.

        Attributes:
            value (Any): The value of the field.
            _register (str | None): An optional register associated with the field.
        """

        value: Any
        _register: str | None

    class Method(TypedDict):
        """
        A TypedDict representing a method with constants and static get objects.

        Attributes:
            consts (List[int]): A list of integer constants used in the method.
            sget_objects (List[SmaliField]): A list of static get objects represented by SmaliField instances.
        """

        consts: List[int]
        sget_objects: List[SmaliField]

    class State(TypedDict):
        """
        State is a TypedDict that represents the state of the deobfuscation process.

        Attributes:
            in_method (bool): Indicates if the current context is within a method.
            in_static_constructor (bool): Indicates if the current context is within a static constructor.
            current_method (SmaliMethod | None): The current method being processed, or None if not applicable.
            registers (Dict[str, register.SmaliRegister]): A dictionary mapping register names to their corresponding SmaliRegister objects.
            calls_to_target_method (List[Any]): A list of calls to the target method.
        """

        in_method: bool
        in_static_constructor: bool
        current_method: SmaliMethod | None
        registers: Dict[str, register.SmaliRegister]
        calls_to_target_method: List[Any]

    def __init__(self, *, filename: str, target_method: SmaliMethod | None = None):
        self.class_name: str | None = None
        self.fields: Dict[SmaliField, ParanoidSmaliParser.Field] = {}
        self.methods: Dict[SmaliMethod, ParanoidSmaliParser.Method] = {}

        self.target_method = target_method
        self._metadata = filename

        self._reset_state()

    def _reset_state(self, key_to_reset: str | None = None):
        """
        Resets the state of the parser.

        If a specific key is provided, only that part of the state will be reset.
        Otherwise, the entire state will be reset to its default values.

        Args:
            key_to_reset (str | None): The specific key in the state to reset. If None, the entire state is reset.
        """
        default_state: ParanoidSmaliParser.State = {
            "in_method": False,
            "current_method": None,
            "in_static_constructor": False,
            "registers": {},
            "calls_to_target_method": [],
        }

        if key_to_reset:
            self.state[key_to_reset] = default_state[key_to_reset]
        else:
            self.state = default_state

    @staticmethod
    def get_fully_qualified_class_name(line: str) -> str:
        """
        Extracts the fully qualified class name from a given Smali class definition line.

        Args:
            line (str): A line from a Smali file that is expected to be a class definition.

        Returns:
            str: The fully qualified class name extracted from the line.

        Raises:
            ParanoidSmaliParserError: If the line does not start with ".class".
        """
        if not line.startswith(".class"):
            raise ParanoidSmaliParserError("Line is not a class definition", {"line": line})

        return line.split()[-1]

    def update(self, line: str, line_number: int = 0):
        """
        Updates the internal state based on the provided Smali code line.

        Args:
            line (str): A single line of Smali code to be processed.

        The method performs the following actions based on the content of the line:
            - Strips leading and trailing whitespace from the line.
            - Skips empty lines.
            - Parses and updates the class name if the line defines a class.
            - Parses and updates fields if the line defines a field.
            - Parses and updates methods if the line defines a method.
            - Updates the state when entering or exiting a method.
            - Updates registers for constant strings and constants.
            - Handles static field gets and puts.
            - Creates arrays and updates their values.
            - Handles static method invocations.

        Raises:
            ParanoidSmaliParserError: If parameters are encountered in an invoke-static instruction.
        """
        line = line.strip()

        # Skip empty lines
        if not line:
            return

        # Get fully qualified class name
        if line.startswith(".class"):
            self.class_name = self.get_fully_qualified_class_name(line)
            return

        # Parse fields
        if line.startswith(".field"):
            field = SmaliField.from_string(line, self.class_name)
            self.fields[field] = {
                "value": None,
                "_register": None,
            }
            return

        # Parse methods
        if line.startswith(".method"):
            method = SmaliMethod.from_string(line, self.class_name)
            self.methods[method] = {
                "consts": [],
                "sget_objects": [],
            }
            self.state["in_method"] = True
            self.state["current_method"] = method
            # logger.debug(f"{line} # -> Entering method")

            # Check if we are in a static constructor
            if "static" in method.modifiers and "constructor" in method.modifiers and method.method == "<clinit>":
                self.state["in_static_constructor"] = True
                # logger.debug(f"{line} # -> Entering static constructor")

            return

        if line.startswith(".end method"):
            self.state["in_method"] = False
            self.state["in_static_constructor"] = False
            # logger.debug(f"{line} # -> Exiting method")
            return

        # Update registers
        if line.startswith("const-string"):
            try:
                instr = instructions.SmaliInstrConstString.parse(line)
            except ValueError:
                return

            # If the register is an array, update the associated field value
            current_value = self.state["registers"].get(instr.register, None)
            if current_value is not None and isinstance(current_value, register.SmaliRegisterArray):
                for field, data in self.fields.items():
                    if data["_register"] == instr.register:
                        self.fields[field]["value"] = self.state["registers"][instr.register].value

            self.state["registers"][instr.register] = register.SmaliRegisterString(instr.value)
            return

        if line.startswith("const"):
            try:
                instr = instructions.SmaliInstrConst.parse(line)
            except ValueError:
                return

            # If the register is an array, update the associated field value
            current_value = self.state["registers"].get(instr.register, None)
            if current_value is not None and isinstance(current_value, register.SmaliRegisterArray):
                for field, data in self.fields.items():
                    if data["_register"] == instr.register:
                        self.fields[field]["value"] = self.state["registers"][instr.register].value

            # Add const to method consts
            if self.state["current_method"]:
                self.methods[self.state["current_method"]]["consts"].append(instr.value)

            self.state["registers"][instr.register] = register.SmaliRegisterConst(instr.value)
            return

        if line.startswith("sget-object"):
            if not self.state["in_method"]:
                return

            try:
                instr = instructions.SmaliInstrSGetSPut.parse(line)
            except ValueError:
                return

            field = SmaliField(instr.field_name, instr.field_type, [], instr.class_name)

            if self.state["current_method"]:
                self.methods[self.state["current_method"]]["sget_objects"].append(field)

        # Create array
        if line.startswith("new-array"):
            try:
                instr = instructions.SmaliInstrNewArray.parse(line)
            except ValueError:
                return

            # Skip array other than [Ljava/lang/String;
            # Necessary for speed
            if instr.type_descriptor != "[Ljava/lang/String;":
                return

            # Get size from register
            if instr.size_register not in self.state["registers"]:
                return

            size_register = self.state["registers"][instr.size_register]
            # Check if size_register is instance of SmaliRegisterConst
            if not isinstance(size_register, register.SmaliRegisterConst):
                logger.debug(self._metadata)
                logger.debug(line)
                logger.debug(f"Size register {instr.size_register} is not a const")
                logger.debug(f" - Class: {size_register.__class__}")
                return

            self.state["registers"][instr.register] = register.SmaliRegisterArray([None] * size_register.get_value())
            return

        # Put value in field
        if line.startswith("sput-object"):
            try:
                instr = instructions.SmaliInstrSGetSPut.parse(line)
            except ValueError:
                return

            # Check if the field is in the current class
            if instr.class_name != self.class_name:
                return

            # Get the field with the matching name
            field = next((f for f in self.fields if f.name == instr.field_name), None)
            if field is None:
                return

            # Save the register and the last value
            self.fields[field]["_register"] = instr.register_dest
            return

        # Insert value in array
        if line.startswith("aput-object"):
            try:
                instr = instructions.SmaliInstrAGetAPut.parse(line)
            except ValueError:
                return

            # Check if arguments are in the registers
            if (
                instr.register_array not in self.state["registers"]
                or instr.register_index not in self.state["registers"]
                or instr.register_dest not in self.state["registers"]
            ):
                return

            array = self.state["registers"][instr.register_array]
            index = self.state["registers"][instr.register_index]
            value = self.state["registers"][instr.register_dest]

            # Check if the array_register is an array
            if not isinstance(array, register.SmaliRegisterArray):
                return

            # Check if the value is a const-string
            if not isinstance(value, register.SmaliRegisterString):
                # logger.warning(f"Value register {instr.register_dest} is not a string")
                # logger.warning(f"{self._metadata}:{line_number+1} -> {line}")
                return

            # Check if the index is a const
            if not isinstance(index, register.SmaliRegisterConst):
                logger.warning(f"Index register {instr.register_index} is not a const")
                return

            # Append the value to the array
            array.value[index.value] = value.value
            return

        if line.startswith("invoke-static"):
            try:
                # Try to parse invoke-static/range first
                instr = instructions.SmaliInstrInvokeStaticRange.parse(line)
                instr = instructions.SmaliInstrInvokeStatic(instr.registers, instr.class_name, instr.method, instr._raw)
            except ValueError:
                try:
                    instr = instructions.SmaliInstrInvokeStatic.parse(line)
                except ValueError:
                    return

            method_name, method_arguments, method_return_type = SmaliMethod.parse_method_signature(instr.method)

            # Check if the target method is the one we are looking for
            if not (
                self.target_method
                and instr.class_name == self.target_method.class_name
                and method_name == self.target_method.method
                and method_arguments == self.target_method.arguments
                and method_return_type == self.target_method.return_type
                and len(instr.registers) == 2
            ):
                return

            first_register = instr.registers[0]

            # TODO: parameters are not supported
            # This is a limitation of the current implementation.
            # It is possible to support them, but it would require a more complex approach.
            if first_register.startswith("p"):
                try:
                    self.state["calls_to_target_method"].append(self.state["registers"][first_register])
                except KeyError:
                    raise ParanoidSmaliParserError(
                        "Parameters are not supported",
                        extra={
                            "registers": self.state["registers"],
                            "register": first_register,
                            "line": line,
                        },
                    )

            try:
                self.state["calls_to_target_method"].append(self.state["registers"][first_register])
            except KeyError:
                raise ParanoidSmaliParserError(
                    "Register not found",
                    extra={
                        "filename": self._metadata,
                        "current_method": self.state["current_method"].to_smali()
                        if self.state["current_method"]
                        else None,
                        "registers": self.state["registers"],
                        "register": first_register,
                        "line": line,
                    },
                )

            return


@overload
def deobfuscate_string(id: int, chunks: List[str]) -> bytes: ...
@overload
def deobfuscate_string(id: int, chunks: List[str], decode: Literal[False]) -> bytes: ...
@overload
def deobfuscate_string(id: int, chunks: List[str], decode: Literal[True]) -> str: ...
def deobfuscate_string(id: int, chunks: List[str], decode: bool = False):
    """
    Deobfuscates a string based on the provided identifier and chunks.

    Args:
        id (int): The identifier for the string to be deobfuscated.
        chunks (List[str]): A list of string chunks used for deobfuscation.
        decode (bool, optional): If True, the result will be decoded to a string. Defaults to False.

    Returns:
        Union[bytes, str]: The deobfuscated string as bytes, or as a decoded string if decode is True.
    """
    result = DeobfuscatorHelper.getString(id, chunks)
    if decode:
        return result.decode()

    return result



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import click

from .commands import deobfuscate, helpers

logger = logging.getLogger(__name__)


@click.group()
@click.option("-v", "--verbose", is_flag=True, help="Enables verbose mode")
def cli(verbose: bool):
    verbosity = logging.DEBUG if verbose else logging.INFO
    log_format = "[%(name)s:%(levelname)s] %(message)s" if verbose else "[%(levelname)s] %(message)s"

    logging.basicConfig(level=verbosity, format=log_format)


cli.add_command(deobfuscate.cli)
cli.add_command(helpers.cli)

if __name__ == "__main__":
    cli()



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import pathlib
import shutil
import sys
from tempfile import NamedTemporaryFile
from typing import Any, Dict, List, TypedDict

import click

from .. import __version__ as deobfuscator_version
from .. import constants, paranoid, report_github_issue_message
from ..encoding import decode_unicode_chunks, encode_smali_string
from ..smali import register

logger = logging.getLogger(__name__)

REMOVED_COMMENT = (
    f"    # Removed with https://github.com/giacomoferretti/paranoid-deobfuscator - v{deobfuscator_version}"
)


class ParanoidSmaliDeobfuscator:
    class SmaliRegisterEncoder(json.JSONEncoder):
        def default(self, o):
            if isinstance(o, register.SmaliRegister):
                return o.to_dict()
            return super().default(o)

    class ParanoidSmaliDeobfuscatorError(Exception):
        def __init__(self, message: str, extra: Dict[str, Any] = {}):
            super().__init__(message)

            self.extra = extra

        def __str__(self):
            if not self.extra:
                return super().__str__()

            return f"{super().__str__()}\n{json.dumps(self.extra, indent=4, cls=ParanoidSmaliDeobfuscator.SmaliRegisterEncoder)}"

    class State(TypedDict):
        class_name: str
        registers: Dict[str, paranoid.register.SmaliRegister]
        last_deobfuscated_string: str | None
        inside_try_block: bool

    def __init__(
        self,
        filepath: pathlib.Path | str,
        target_method: paranoid.SmaliMethod,
        obfuscated_chunks: List[str],
        # edit_in_place: bool = True,
    ):
        self.filepath = filepath
        # if edit_in_place:
        #     self.file = open(filepath, "r+")
        # else:
        #     self.file = open(filepath, "r")
        self.file = open(filepath, "r")
        self.tmp_file = NamedTemporaryFile(mode="wt", dir=pathlib.Path(filepath).parent.absolute(), delete=False)

        self.target_method = target_method
        self.obfuscated_chunks = obfuscated_chunks
        self._reset_state()

    def _reset_state(self, key_to_reset: str | None = None):
        default_state: ParanoidSmaliDeobfuscator.State = {
            "class_name": "",
            "registers": {},
            "last_deobfuscated_string": None,
            "inside_try_block": False,
        }

        if key_to_reset:
            self.state[key_to_reset] = default_state[key_to_reset]
        else:
            self.state = default_state

    @staticmethod
    def get_fully_qualified_class_name(line: str) -> str:
        if not line.startswith(".class"):
            raise Exception("Line does not start with .class")

        return line.split()[-1]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.file.close()

    def process(self, line: str, line_num: int) -> str | None:
        # Reset the state if the result from the getString method is not used
        if self.state["last_deobfuscated_string"] and (line.startswith("invoke") or line.startswith("if-")):
            self.state["last_deobfuscated_string"] = None

        if line.startswith(":try_start"):
            self.state["inside_try_block"] = True
            return

        if line.startswith(":try_end"):
            self.state["inside_try_block"] = False
            return

        # Get fully qualified class name
        if line.startswith(".class"):
            self.state["class_name"] = self.get_fully_qualified_class_name(line)
            return

        # Update registers
        if line.startswith("const"):
            try:
                instr = paranoid.instructions.SmaliInstrConst.parse(line)
            except ValueError:
                return

            self.state["registers"][instr.register] = paranoid.register.SmaliRegisterConst(instr.value)
            return

        # Search for calls to the target method
        if line.startswith("invoke-static"):
            try:
                # Try to parse invoke-static/range first
                instr = paranoid.instructions.SmaliInstrInvokeStaticRange.parse(line)
                instr = paranoid.instructions.SmaliInstrInvokeStatic(
                    instr.registers, instr.class_name, instr.method, instr._raw
                )
            except ValueError:
                try:
                    instr = paranoid.instructions.SmaliInstrInvokeStatic.parse(line)
                except ValueError:
                    return

            method_name, method_arguments, method_return_type = paranoid.SmaliMethod.parse_method_signature(
                instr.method
            )

            # Check if the target method is the one we are looking for
            if not (
                self.target_method
                and instr.class_name == self.target_method.class_name
                and method_name == self.target_method.method
                and method_arguments == self.target_method.arguments
                and method_return_type == self.target_method.return_type
                and len(instr.registers) == 2
            ):
                return

            first_register = instr.registers[0]

            # Get the value of the register
            register_value = self.state["registers"].get(first_register)

            # TODO: parameters are not supported
            # This is a limitation of the current implementation.
            # It is possible to support them, but it would require a more complex approach.
            if first_register.startswith("p") and not register_value:
                raise ParanoidSmaliDeobfuscator.ParanoidSmaliDeobfuscatorError(
                    "Parameters are not supported",
                    extra={
                        "registers": self.state["registers"],
                        "register": first_register,
                        "line": line,
                    },
                )

            if not register_value:
                raise ParanoidSmaliDeobfuscator.ParanoidSmaliDeobfuscatorError(
                    "Register not found",
                    extra={
                        "registers": self.state["registers"],
                        "register": first_register,
                        "line": line,
                    },
                )

            # Check if the register is a constant
            if not isinstance(register_value, paranoid.register.SmaliRegisterConst):
                raise ParanoidSmaliDeobfuscator.ParanoidSmaliDeobfuscatorError(
                    "Register is not a constant",
                    extra={
                        "registers": self.state["registers"],
                        "register": first_register,
                        "line": line,
                    },
                )

            # Deobfuscate the string
            deobfuscated_string = paranoid.deobfuscate_string(register_value.value, self.obfuscated_chunks, True)
            self.state["last_deobfuscated_string"] = deobfuscated_string

            # Edge case: if we are inside a try block, we need to have at least one valid instruction,
            # so we add a nop instruction, otherwise the smali file will be invalid and the APK will recompile
            if self.state["inside_try_block"]:
                return f"    nop {REMOVED_COMMENT.strip()}"

            return REMOVED_COMMENT

        # Move result object
        if line.startswith("move-result-object"):
            try:
                instr = paranoid.instructions.SmaliInstrMoveResult.parse(line)
            except ValueError:
                return

            if self.state["last_deobfuscated_string"] is not None:
                new_line = f'    const-string {instr.register}, "{encode_smali_string(self.state["last_deobfuscated_string"])}"'
                self.state["last_deobfuscated_string"] = None
                return new_line

            return

        return

    def update(self, _line: str, line_num: int = 0):
        line = _line.strip()

        # Skip empty lines
        if not line:
            self.tmp_file.write(_line)
            return

        # Process the line
        try:
            updated_line = self.process(line, line_num)
            if updated_line is not None:
                self.tmp_file.write(updated_line + "\n")
                return
        except ParanoidSmaliDeobfuscator.ParanoidSmaliDeobfuscatorError as e:
            # Ignore Parameters are not supported
            if e.args[0] == "Parameters are not supported":
                logger.warning(f"{self.filepath}:{line_num+1}: Detected unsupported method call")
                # Add the line to the temporary file
                self.tmp_file.write(_line)
                return

            # Log and raise the error
            logger.error(report_github_issue_message(str(e)))
            raise e

        # Add the line to the temporary file
        self.tmp_file.write(_line)


@click.command(name="deobfuscate", help="Deobfuscate a paranoid obfuscated APK smali files")
@click.argument("target", type=click.Path(exists=True, file_okay=False))
def cli(target: str):
    target_directory = pathlib.Path(target)

    # First pass: find the get string method and the obfuscated string array
    potential_get_string_methods = []
    potential_obfuscated_string_arrays = []
    for smali_file in target_directory.rglob("*.smali"):
        with open(smali_file, "r") as f:
            smali_parser = paranoid.ParanoidSmaliParser(filename=str(smali_file))

            for line_num, line in enumerate(f):
                smali_parser.update(line, line_num)

            # Add potential get string methods
            for method, data in smali_parser.methods.items():
                if (
                    data["consts"] == constants.PARANOID_GET_STRING_CONST_SIGNATURE
                    and method.arguments == constants.PARANOID_GET_STRING_ARGUMENTS
                    and method.return_type == constants.PARANOID_GET_STRING_RETURN_TYPE
                ):
                    potential_get_string_methods.append((method, data["sget_objects"]))

            # Add potential obfuscated string arrays
            for field, data in smali_parser.fields.items():
                if field.type == "[Ljava/lang/String;":
                    potential_obfuscated_string_arrays.append((field, data["value"]))

    # Check if only one method is found
    if len(potential_get_string_methods) != 1:
        logger.error("Found more than one potential get string method")
        logger.error("This is not supported yet")
        sys.exit(1)

    get_string_method, get_string_fields = potential_get_string_methods[0]
    get_string_field: paranoid.SmaliField = get_string_fields[0]

    # Check if only one field is found
    if len(get_string_fields) != 1:
        logger.error("Found more than one potential obfuscated string array")
        logger.error("This is not supported yet")
        sys.exit(1)

    # Extract the string chunks
    chunks = []
    for field, value in potential_obfuscated_string_arrays:
        if field.class_name == get_string_field.class_name and field.name == get_string_field.name:
            chunks = value

    # Check if the chunks are found
    if not chunks:
        logger.error("No chunks found")
        return

    logger.debug(f"Method: {get_string_method}")
    logger.debug(f"Field: {get_string_field}")
    logger.debug("Chunks:")
    logger.debug(chunks)

    # Decode the chunks
    chunks = decode_unicode_chunks(chunks)

    # Second pass: deobfuscate file
    for smali_file in target_directory.rglob("*.smali"):
        with ParanoidSmaliDeobfuscator(smali_file, get_string_method, chunks) as deobfuscator:
            for line_num, line in enumerate(deobfuscator.file):
                deobfuscator.update(line, line_num)

        # Replace the original file with the temporary one
        shutil.move(deobfuscator.tmp_file.name, smali_file)



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC, abstractmethod


class SmaliRegister(ABC):
    def __init__(self, value):
        self.value = value
        self.validate_value(value)

    @abstractmethod
    def validate_value(self, value):
        """Validate the value for the specific type."""
        pass

    def __repr__(self):
        return f"{self.__class__.__name__}(value={self.value})"

    def get_value(self):
        return self.value

    @classmethod
    @abstractmethod
    def get_type(cls) -> str:
        """Return the type of the register."""
        pass

    def to_dict(self):
        return {"type": self.get_type(), "value": self.value}


class SmaliRegisterConst(SmaliRegister):
    def validate_value(self, value):
        if not isinstance(value, int):
            raise ValueError("For 'const', value must be an int")

    @classmethod
    def get_type(cls):
        return "const"


class SmaliRegisterArray(SmaliRegister):
    def validate_value(self, value):
        if not isinstance(value, list):
            raise ValueError("For 'array', value must be a list")

    @classmethod
    def get_type(cls):
        return "array"


class SmaliRegisterString(SmaliRegister):
    def validate_value(self, value):
        if not isinstance(value, str):
            raise ValueError("For 'string', value must be a string")

    @classmethod
    def get_type(cls):
        return "string"



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version_major__ = 3
__version_minor__ = 0
__version_patch__ = 1
__version_metadata__ = ""
__version__ = f"{__version_major__}.{__version_minor__}.{__version_patch__}{__version_metadata__}"
__issue_tracker__ = "https://github.com/giacomoferretti/paranoid-deobfuscator/issues"


def report_github_issue_message(data: str | None = None) -> str:  # pragma: no cover
    result = f"Please report this issue at {__issue_tracker__}"
    if data:
        result += f"\n\nReport information:\n{data}"
    return result
